
void locBlock(int* l_lb, int* l_ub, int* l_stride,
	      int g_lb, int g_ub, int offset, int stride,
	      int t_lb, int t_ub, int distCode, int numProcs, int coord) {

  /**************************************************************************/
  /*                                                                        */
  /* Define `l_lb', `l_ub', and `l_stride' for specified value of `coord'   */
  /*                                                                        */
  /**************************************************************************/

  int g_extent  = g_ub - g_lb + 1 ;
  int base      = offset ;             /* we think */

  switch(distCode) {
    case 1 : {

      /********************************************************/
      /*                                                      */
      /*  BLOCK distribution                                  */
      /*                                                      */
      /********************************************************/

      int t_blockSize = (t_ub - t_lb + numProcs) / numProcs ;

      int blockStart = coord * t_blockSize ;

      int blockStartOffset  = blockStart - base ;
      int blockFinishOffset = blockStartOffset + t_blockSize - 1 ;

      if(stride > 0) {
	int i_lo = (blockStartOffset + stride - 1) / stride ;
	int i_hi = blockFinishOffset / stride ;

	if(i_lo < 0)         i_lo = 0 ;
	if(i_hi >= g_extent) i_hi = g_extent - 1 ;

	*l_lb = stride * i_lo - blockStartOffset + 1 ;
	*l_ub = stride * i_hi - blockStartOffset + 1 ;
        *l_stride = stride ;
      }
      else {
	int i_lo = (blockFinishOffset + stride + 1) / stride ;
	int i_hi = blockStartOffset / stride ;

	if(i_lo < 0)         i_lo = 0 ;
	if(i_hi >= g_extent) i_hi = g_extent - 1 ;

	*l_lb = stride * i_hi - blockStartOffset + 1 ;
	*l_ub = stride * i_lo - blockStartOffset + 1 ;
        *l_stride = -stride ;
      }

      break ;
    }
    case 2 : {

      /********************************************************/
      /*                                                      */
      /*  CYCLIC distribution                                 */
      /*                                                      */
      /********************************************************/

      int c = coord - base ;
      int f, f_s ;
      int g, g_s ;

      /*
       * Compute `f', the GCD of `numProcs' and `stride',
       * and `f_s', a solution of
       *
       *     f_s * s = f  mod  p
       *
       * (s = stride, p = numProcs) 
       *
       */

      f   = numProcs ;
      f_s = 0 ;
      if(stride > 0) {
	g   = stride ;
	g_s = 1 ;
      }
      else {
	g   = -stride ;
	g_s = -1 ;
      }
      while(g) {
	int temp ;
	int r = f / g ;

	temp = g ;
	g    = f - r * g ;
	f    = temp ;
      
	temp = g_s ;
	g_s  = f_s - r * g_s ;
	f_s  = temp ;
      }

      if(c % f == 0) {

	/*
	 * Compute the smallest and largest solutions of
	 *
	 *     base + i * s = coord  mod p
	 *
	 * in the range 0, ..., g_extent - 1, viz
	 *
	 *     i_lo = a + b * ceiling(- a / b)
	 *     i_hi = a + b * floor((g_extent - 1 - a) / b)
	 *
	 * where a = c / f * f_s, b = p / f, c = coord - base.
	 *
	 * Solutions exist iff `f' divides `c'.
	 *
	 */

	int a = (c / f) * f_s, b = numProcs / f ;
        int i_lo, i_hi ;

	if(a >= 0)
	  i_lo = a - b * (a / b) ;
	else
	  i_lo = a + b * ((-a + b - 1) / b) ;

	if(g_extent > a)
	  i_hi = a + b * ((g_extent - 1 - a) / b) ;
	else
	  i_hi = a - b * ((- g_extent + a + b) / b) ;

	/*
	 * Finally, define local subscript range.
	 *
	 */

	if(stride > 0) {
	  *l_lb = (base + stride * i_lo) / numProcs + 1 ;
	  *l_ub = (base + stride * i_hi) / numProcs + 1 ;
	  *l_stride = stride / f ;
	}
	else {
	  *l_lb = (base + stride * i_hi) / numProcs + 1 ;
	  *l_ub = (base + stride * i_lo) / numProcs + 1 ;
	  *l_stride = - stride / f ;
	}
      }
      else {

	/* No solution */

	*l_lb = 1 ;
	*l_ub = 0 ;
	*l_stride = 1 ;
      }

      break ;
    }
    case 3 : {

      /********************************************************/
      /*                                                      */
      /*  Collapsed (*) distribution                          */
      /*                                                      */
      /********************************************************/

      if(stride > 0) {
	*l_lb = base ;
	*l_ub = base + stride * g_extent ;
        *l_stride = stride ;
      }
      else {
	*l_lb = base + stride * g_extent ;
	*l_ub = base ;
        *l_stride = -stride ;
      }

      break ;
    }
  }
}

void locToGlob(int* g_i, int coord, int l_i,
	       int g_lb, int g_ub, int offset, int stride,
	       int t_lb, int t_ub, int distCode, int numProcs) {

  /*************************************************************************/
  /*                                                                       */
  /* In the context of a given set of range parameters,                    */
  /* translate a process coordinate `coord' and a local subscript `l_i'    */
  /* to a global subscript `g_i'.                                          */
  /*                                                                       */
  /*************************************************************************/

  int g_extent    = g_ub - g_lb + 1 ;

  switch(distCode) {
    case 1 : {

      int t_blockSize = (t_ub - t_lb + numProcs) / numProcs ;
      int i_t = coord * t_blockSize + l_i - 1 ;
      *g_i = g_lb + (i_t - offset) / stride ;

      break ;
    }
    case 2 : {

      int i_t = coord + (l_i - 1) * numProcs ;
      *g_i = g_lb + (i_t - offset) / stride ;

      break ;
    }
    case 3 : {

      int i_t = l_i - 1 ;
      *g_i = g_lb + (i_t - offset) / stride ;

      break ;
    }
  }
}

void main(int argc, char** argv) {
  
  int t_lb, t_ub, distCode, numProcs ;
  int g_lb, g_ub, offset, stride ;

  int coord ;
  int t_blockSize ;
  int *map ;
  int l_i ;

  printf("Distribution format?: ") ;
  scanf("%d", &distCode) ;
  switch(distCode) {
    case 1 : {
      printf("BLOCK Distribution.\n") ;
      printf("Number of processors?: ") ;
      scanf("%d", &numProcs) ;

      break ;
    }
    case 2 : {
      printf("CYCLIC Distribution.\n") ;
      printf("Number of processors?: ") ;
      scanf("%d", &numProcs) ;

      break ;
    }
    case 3 : {
      printf("Collapsed Distribution.\n") ;
      numProcs = 1 ;
      printf("Taking number of processors = %d\n", numProcs) ;
      break ;
    }
    default : {
      printf("Unrecognised\n") ;
      exit(1) ;
    }
  }

  printf("Template extent?: ") ;
  scanf("%d", &t_ub) ;
  t_lb = 1 ;
  printf("Taking template template lower bound = %d\n", t_lb) ;

  printf("Array lower bound?: ") ;
  scanf("%d", &g_lb) ;
  printf("Array upper bound?: ") ;
  scanf("%d", &g_ub) ;
  printf("Alignment offset?: ") ;
  scanf("%d", &offset) ;
  printf("Alignment stride?: ") ;
  scanf("%d", &stride) ;
  printf("\n") ;

  t_blockSize = (t_ub - t_lb + numProcs) / numProcs ;
  map = (int*) malloc(sizeof(int) * numProcs * t_blockSize) ;

  printf("Local block parameters:\n\n") ;
  printf("%10s%10s%10s%10s\n", "coord", "l_lb", "l_ub", "l_stride") ;

  for(coord = 0 ; coord < numProcs ; coord++) {
    int l_lb, l_ub, l_stride ;

    for(l_i = 1 ; l_i <= t_blockSize ; l_i++)
      map [t_blockSize * coord + l_i - 1] = g_lb - 1 ;

    locBlock(&l_lb, &l_ub, &l_stride,
	     g_lb, g_ub, offset, stride,
	     t_lb, t_ub, distCode, numProcs, coord) ;

    printf("%10d%10d%10d%10d\n", coord, l_lb, l_ub, l_stride) ;

    for(l_i = l_lb ; l_i <= l_ub ; l_i += l_stride)
      locToGlob(&(map [t_blockSize * coord + l_i - 1]),
                coord, l_i,
	        g_lb, g_ub, offset, stride,
	        t_lb, t_ub, distCode, numProcs) ;
  }
  printf("\n") ;

  printf("Mapping of global index values:\n\n") ;
  printf("l_i\\coord") ;
  printf("%5d", 0) ;
  for(coord = 1 ; coord < numProcs ; coord++)
    printf("%8d", coord) ;
  printf("\n\n") ;
  for(l_i = 1 ; l_i <= t_blockSize ; l_i++) {
    printf("%4d  ", l_i) ;
    for(coord = 0 ; coord < numProcs ; coord++) {
      int g_i = map [t_blockSize * coord + l_i - 1] ;
      if(g_i >= g_lb)
        printf("%8d", g_i) ;
      else
        printf("       -") ;
    }
    printf("\n\n") ;
  }
}

